@@ -59,6 +59,7 @@ boot_devices = {
     'optical': 0x14,
     'dvd': 0x14,
     'floppy': 0x3c,
+    'usb': 0x3c,
     'default': 0x0,
     'setup': 0x18,
     'bios': 0x18,
@@ -1285,6 +1286,8 @@ class Command(object):
         """Set the management controller identifier, per DCMI specification
 
         """
+        if not isinstance(mci, bytes):
+            mci = mci.encode('utf8')
         return self._chunkwise_dcmi_set(0xa, mci + b'\x00')
 
     def get_asset_tag(self):
@@ -1322,7 +1325,7 @@ class Command(object):
         chunks = [data[i:i + 15] for i in range(0, len(data), 15)]
         offset = 0
         for chunk in chunks:
-            chunk = bytearray(chunk, 'utf-8')
+            chunk = bytearray(chunk)
             cmddata = bytearray((0xdc, offset, len(chunk)))
             cmddata += chunk
             self.xraw_command(netfn=0x2c, command=command, data=cmddata)
@@ -2025,7 +2028,7 @@ class Command(object):
         self.oem_init()
         return self._oem.detach_remote_media()
 
-    def upload_media(self, filename, progress=None):
+    def upload_media(self, filename, progress=None, data=None):
         """Upload a file to be hosted on the target BMC
 
         This will upload the specified data to
@@ -2037,7 +2040,7 @@ class Command(object):
         :param progress: Optional callback for progress updates
         """
         self.oem_init()
-        return self._oem.upload_media(filename, progress)
+        return self._oem.upload_media(filename, progress, data)
 
     def list_media(self):
         """List attached remote media
@@ -2063,6 +2066,6 @@ class Command(object):
         self.oem_init()
         return self._oem.save_licenses(directory)
 
-    def apply_license(self, filename, progress=None):
+    def apply_license(self, filename, progress=None, data=None):
         self.oem_init()
-        return self._oem.apply_license(filename, progress)
+        return self._oem.apply_license(filename, progress, data)
