@@ -165,7 +165,6 @@ class OEMHandler(generic.OEMHandler):
             self.tsmahandler.set_credentials(
                 ipmicmd.ipmi_session.userid.decode('utf-8'),
                 ipmicmd.ipmi_session.password.decode('utf-8'))
-            self.tsmahandler.init_redfish()
 
     @property
     def _megarac_eth_index(self):
@@ -285,6 +284,8 @@ class OEMHandler(generic.OEMHandler):
     def reseat_bay(self, bay):
         if self.is_fpc:
             return self.smmhandler.reseat_bay(bay)
+        elif self.has_xcc and bay == -1:
+            return self.immhandler.reseat()
         return super(OEMHandler, self).reseat_bay(bay)
 
     def get_ntp_enabled(self):
@@ -293,6 +294,8 @@ class OEMHandler(generic.OEMHandler):
             return ntpres['data'][0] == '\x01'
         elif self.is_fpc:
             return self.smmhandler.get_ntp_enabled(self._fpc_variant)
+        elif self.has_tsma:
+            return self.tsmahandler.get_ntp_enabled()
         return None
 
     def get_ntp_servers(self):
@@ -304,6 +307,8 @@ class OEMHandler(generic.OEMHandler):
             return srvs
         if self.is_fpc:
             return self.smmhandler.get_ntp_servers()
+        if self.has_tsma:
+            return self.tsmahandler.get_ntp_servers()
         return None
 
     def set_ntp_enabled(self, enabled):
@@ -318,6 +323,8 @@ class OEMHandler(generic.OEMHandler):
         if self.is_fpc:
             self.smmhandler.set_ntp_enabled(enabled)
             return True
+        if self.has_tsma:
+            self.tsmahandler.set_ntp_enabled(enabled)
         return None
 
     def set_ntp_server(self, server, index=0):
@@ -334,6 +341,10 @@ class OEMHandler(generic.OEMHandler):
                     'SMM supports indexes 0 through 2')
             self.smmhandler.set_ntp_server(server, index)
             return True
+        elif self.has_tsma:
+            if not (0 <= index <= 1):
+                raise pygexc.InvalidParameterValue("Index must be 0 or 1")
+            return self.tsmahandler.set_ntp_server(server, index)
         return None
 
     def set_user_access(self, uid, channel, callback, link_auth, ipmi_msg,
@@ -351,12 +362,18 @@ class OEMHandler(generic.OEMHandler):
             return self._fpc_variant
         fpc_ids = ((19046, 32, 1063), (20301, 32, 462))
         smm_id = (19046, 32, 1180)
+        smmv2_2u = (19046, 32, 1360)
+        smmv2_6u = (19046, 32, 1387)
         currid = (self.oemid['manufacturer_id'], self.oemid['device_id'],
                   self.oemid['product_id'])
         if currid in fpc_ids:
             self._fpc_variant = 6
         elif currid == smm_id:
             self._fpc_variant = 2
+        elif currid == smmv2_2u:
+            self._fpc_variant = 0x22
+        elif currid == smmv2_6u:
+            self._fpc_variant = 0x26
         return self._fpc_variant
 
     @property
@@ -577,7 +594,7 @@ class OEMHandler(generic.OEMHandler):
             if self.has_xcc and name and name.startswith('PSU '):
                 self.immhandler.augment_psu_info(fru, name)
             return fru
-        elif self.is_fpc == 2:  # SMM variant
+        elif self.is_fpc and self.is_fpc != 6:  # SMM variant
             fru['oem_parser'] = 'lenovo'
             return self.smmhandler.process_fru(fru)
         else:
@@ -635,8 +652,8 @@ class OEMHandler(generic.OEMHandler):
             return nextscale.get_fpc_firmware(bmcver, self.ipmicmd,
                                               self._fpc_variant)
         elif self.has_tsma:
-            return self.tsmahandler.get_firmware_inventory(components,
-                                                           raisebypass=False)
+            return self.tsmahandler.get_firmware_inventory(
+                components, raisebypass=False, ipmicmd=self.ipmicmd)
         return super(OEMHandler, self).get_oem_firmware(bmcver, components)
 
     def get_diagnostic_data(self, savefile, progress, autosuffix=False):
@@ -645,7 +662,8 @@ class OEMHandler(generic.OEMHandler):
                                                        autosuffix)
         if self.is_fpc:
             return self.smmhandler.get_diagnostic_data(savefile, progress,
-                                                       autosuffix)
+                                                       autosuffix,
+                                                       self._fpc_variant)
         if self.has_tsma:
             return self.tsmahandler.get_diagnostic_data(savefile, progress,
                                                         autosuffix)
@@ -980,6 +998,8 @@ class OEMHandler(generic.OEMHandler):
             return self.immhandler.get_bmc_configuration()
         if self.is_fpc:
             return self.smmhandler.get_bmc_configuration()
+        if self.has_tsma:
+            return self.tsmahandler.get_bmc_configuration()
         return super(OEMHandler, self).get_bmc_configuration()
 
     def set_bmc_configuration(self, changeset):
@@ -988,6 +1008,9 @@ class OEMHandler(generic.OEMHandler):
         if self.is_fpc:
             return self.smmhandler.set_bmc_configuration(
                 changeset, self._fpc_variant)
+        if self.has_tsma:
+            return self.tsmahandler.set_bmc_configuration(
+                changeset)
         return super(OEMHandler, self).set_bmc_configuration(changeset)
 
     def get_system_configuration(self, hideadvanced):
@@ -1028,10 +1051,10 @@ class OEMHandler(generic.OEMHandler):
                 netfn=0x32, command=0x9f, data=(8, 10, 0, 0))
             self.ipmicmd.xraw_command(netfn=0x32, command=0x9f, data=(8, 11))
 
-    def upload_media(self, filename, progress):
+    def upload_media(self, filename, progress, data):
         if self.has_xcc or self.has_imm:
-            return self.immhandler.upload_media(filename, progress)
-        return super(OEMHandler, self).upload_media(filename, progress)
+            return self.immhandler.upload_media(filename, progress, data)
+        return super(OEMHandler, self).upload_media(filename, progress, data)
 
     def list_media(self):
         if self.has_xcc or self.has_imm:
@@ -1065,7 +1088,7 @@ class OEMHandler(generic.OEMHandler):
             return self.immhandler.save_licenses(directory)
         return super(OEMHandler, self).save_licenses(directory)
 
-    def apply_license(self, filename, progress=None):
+    def apply_license(self, filename, progress=None, data=None):
         if self.has_xcc:
-            return self.immhandler.apply_license(filename, progress)
-        return super(OEMHandler, self).apply_license(filename, progress)
+            return self.immhandler.apply_license(filename, progress, data)
+        return super(OEMHandler, self).apply_license(filename, progress, data)
