@@ -127,8 +127,7 @@ class IMMClient(object):
         srv = self.imm
         if ':' in srv:
             srv = '[{0}]'.format(srv)
-        self.adp_referer = 'https://{0}/designs/imm/index-console.php'.format(
-            srv)
+        self.adp_referer = 'https://imm/designs/imm/index-console.php'
         if ipmicmd.ipmi_session.password:
             self.username = ipmicmd.ipmi_session.userid.decode('utf-8')
             self.password = ipmicmd.ipmi_session.password.decode('utf-8')
@@ -254,6 +253,15 @@ class IMMClient(object):
                 else:
                     raise pygexc.InvalidParameterValue(
                         '{0} not a known setting'.format(key))
+        self.merge_changeset(changeset)
+        if changeset:
+            try:
+                self.fwc.set_fw_options(self.fwo)
+            finally:
+                self.fwo = None
+                self.fwovintage = 0
+
+    def merge_changeset(self, changeset):
         for key in changeset:
             if isinstance(changeset[key], six.string_types):
                 changeset[key] = {'value': changeset[key]}
@@ -294,12 +302,6 @@ class IMMClient(object):
                 self.fwo[key]['new_value'] = newnewvalues[0]
             else:
                 self.fwo[key]['new_value'] = newnewvalues
-        if changeset:
-            try:
-                self.fwc.set_fw_options(self.fwo)
-            except Exception:
-                self.fwo = None
-                raise
 
     def clear_bmc_configuration(self):
         self.ipmicmd.xraw_command(0x2e, 0xcc,
@@ -354,8 +356,9 @@ class IMMClient(object):
                            'password': self.password,
                            'SessionTimeout': 60})
         headers = {'Connection': 'keep-alive',
-                   'Referer': 'https://{0}/designs/imm/index.php'.format(
-                       self.imm),
+                   'Origin': 'https://imm/',
+                   'Host': 'imm',
+                   'Referer': 'https://imm/designs/imm/index.php',
                    'Content-Type': 'application/x-www-form-urlencoded'}
         wc.request('POST', '/data/login', adata, headers)
         rsp = wc.getresponse()
@@ -371,7 +374,8 @@ class IMMClient(object):
                 else:
                     self.uploadtoken = {}
                 wc.set_header('Referer', self.adp_referer)
-                wc.set_header('Host', self.imm)
+                wc.set_header('Host', 'imm')
+                wc.set_header('Origin', 'https://imm/')
                 return wc
 
     @property
@@ -404,7 +408,12 @@ class IMMClient(object):
     def grab_cacheable_json(self, url, age=30):
         data = self.get_cached_data(url, age)
         if not data:
-            data = self.wc.grab_json_response(url)
+            data, status = self.wc.grab_json_response_with_status(url)
+            if status == 401:
+                self._wc = None
+                data, status = self.wc.grab_json_response_with_status(url)
+            if status != 200:
+                data = {}
             self.datacache[url] = (data, util._monotonic_time())
         return data
 
@@ -416,7 +425,7 @@ class IMMClient(object):
         except KeyError:
             return None
 
-    def upload_media(self, filename, progress=None):
+    def upload_media(self, filename, progress=None, data=None):
         xid = random.randint(0, 1000000000)
         alloc = self.wc.grab_json_response(
             '/data/set',
@@ -430,7 +439,7 @@ class IMMClient(object):
         uploadfields['available'] = alloc['available']
         uploadfields['checksum'] = xid
         ut = webclient.FileUploader(
-            self.wc, '/designs/imm/upload/rp_image_upload.esp', filename,
+            self.wc, '/designs/imm/upload/rp_image_upload.esp', filename, data,
             otherfields=uploadfields)
         ut.start()
         while ut.isAlive():
@@ -481,6 +490,11 @@ class IMMClient(object):
                 for uload in item['images']:
                     if uload['status'] != 0:
                         yield media.Media(uload['filename'])
+            for attached in item.get('urls', []):
+                filename = attached['url']
+                filename = filename.split('/')[-1]
+                url = '/'.join(attached['url'].split('/')[:-1])
+                yield media.Media(filename, url)
 
     def detach_remote_media(self):
         mnt = self.wc.grab_json_response(
@@ -846,6 +860,13 @@ class XCCClient(IMMClient):
         super(XCCClient, self).__init__(ipmicmd)
         self.adp_referer = None
 
+    def reseat(self):
+        rsp = self.wc.grab_json_response_with_status(
+            '/api/providers/virt_reseat', '{}')
+        if rsp[1] != 200 or rsp[0].get('return', 1) != 0:
+            raise pygexc.UnsupportedFunctionality(
+                'This platform does not support AC reseat.')
+
     def get_description(self):
         dsc = self.wc.grab_json_response('/DeviceDescription.json')
         dsc = dsc[0]
@@ -881,20 +902,29 @@ class XCCClient(IMMClient):
             'value': passrules.get('pass_min_length')}
         settings['password_lockout_period'] = {
             'value': passrules.get('lockout_period')}
+        presassert = self.wc.grab_json_response('/api/dataset/imm_rpp')
+        asserted = presassert.get('items', [{}])[0].get('rpp_Assert', None)
+        if asserted is not None:
+            settings['presence_assert'] = {
+                'value': 'Enable' if asserted else 'Disable'
+            }
         try:
             enclosureinfo = self.ipmicmd.xraw_command(0x3a, 0xf1, data=[0])
         except pygexc.IpmiException:
             return settings
         settings['smm'] = {
             'default': 'Disable',
-            'possible': ['Enable', 'Disable'],
-            'help': 'Enables or disables the network of the D2 '
-                    'enclosure manager.',
+            'possible': ['Enable', 'Disable', 'IPMI'],
+            'help': 'Enables or disables the network of the '
+                    'enclosure manager. IPMI Enables with IPMI '
+                    'for v2 systems.',
         }
         if bytearray(enclosureinfo['data'])[0] == 2:
             settings['smm']['value'] = 'Disable'
         elif bytearray(enclosureinfo['data'])[0] == 1:
             settings['smm']['value'] = 'Enable'
+        elif bytearray(enclosureinfo['data'])[0] == 4:
+            settings['smm']['value'] = 'IPMI'
         else:
             settings['smm']['value'] = None
         smmip = self.get_property('/v2/ibmc/smm/smm_ip')
@@ -923,15 +953,31 @@ class XCCClient(IMMClient):
                 changeset[key] = {'value': changeset[key]}
             currval = changeset[key].get('value', None)
             if 'smm'.startswith(key.lower()):
-                if currval == 'Enable':
+                if 'enabled'.startswith(currval.lower()):
                     self.ipmicmd.xraw_command(0x3a, 0xf1, data=[1])
-                elif currval == 'Disable':
+                elif 'disabled'.startswith(currval.lower()):
                     self.ipmicmd.xraw_command(0x3a, 0xf1, data=[2])
+                elif 'ipmi'.startswith(currval.lower()):
+                    self.ipmicmd.xraw_command(0x3a, 0xf1, data=[4])
             elif key.lower() in self.rulemap:
                 ruleset[self.rulemap[key.lower()]] = changeset[key]['value']
                 if key.lower() == 'password_expiration':
                     warntime = str(int(int(changeset[key]['value']) * 0.08))
                     ruleset['USER_GlobalPassExpWarningPeriod'] = warntime
+            elif 'presence_asserted'.startswith(key.lower()):
+                assertion = changeset[key]['value']
+                if 'enabled'.startswith(assertion.lower()):
+                    self.wc.grab_json_response('/api/dataset',
+                                               {'IMM_RPPAssert': '0'})
+                    self.wc.grab_json_response('/api/dataset',
+                                               {'IMM_RPPAssert': '1'})
+                elif 'disabled'.startswith(assertion.lower()):
+                    self.wc.grab_json_response('/api/dataset',
+                                               {'IMM_RPPAssert': '0'})
+                else:
+                    raise pygexc.InvalidParameterValue(
+                        '"{0}" is not a recognized value for {1}'.format(
+                            assertion, key))
             else:
                 raise pygexc.InvalidParameterValue(
                     '{0} not a known setting'.format(key))
@@ -944,7 +990,8 @@ class XCCClient(IMMClient):
             {'Action': 'Bios.ResetBios'},
             headers={
                 'Authorization': 'Basic %s' % base64.b64encode(
-                    self.username + ':' + self.password),
+                    (self.username + ':' + self.password).encode('utf8')
+                ).decode('utf8'),
                 'Content-Type': 'application/json'
             }
         )
@@ -1039,7 +1086,7 @@ class XCCClient(IMMClient):
                                                 {'Generate_FFDC_status': 1})
         url = '/ffdc/{0}'.format(result['FileName'])
         if autosuffix and not savefile.endswith('.tzz'):
-            savefile += '.tzz'
+            savefile += '-{0}'.format(result['FileName'])
         fd = webclient.FileDownloader(self.wc, url, savefile)
         fd.start()
         while fd.isAlive():
@@ -1285,10 +1332,18 @@ class XCCClient(IMMClient):
                     nameappend += 1
             else:
                 name = vol.name
-            if vol.stripsize:
+            if vol.stripsize is not None:
                 stripsize = int(math.log(vol.stripsize * 2, 2))
             else:
                 stripsize = props['stripsize']
+            if vol.read_policy is not None:
+                read_policy = vol.read_policy
+            else:
+                read_policy = props["cpra"]
+            if vol.write_policy is not None:
+                write_policy = vol.write_policy
+            else:
+                write_policy = props["cpwb"]
             strsize = 'remainder' if vol.size is None else str(vol.size)
             if strsize in ('all', '100%'):
                 volsize = params['capacity']
@@ -1308,7 +1363,7 @@ class XCCClient(IMMClient):
                 raise pygexc.InvalidParameterValue(
                     'Requested sizes exceed available capacity')
             vols.append('{0};{1};{2};{3};{4};{5};{6};{7};{8};|'.format(
-                name, volsize, stripsize, props['cpwb'], props['cpra'],
+                name, volsize, stripsize, write_policy, read_policy,
                 props['cpio'], props['ap'], props['dcp'], props['initstate']))
         url = '/api/function'
         cid = params['controller'].split(',')
@@ -1636,11 +1691,11 @@ class XCCClient(IMMClient):
                 yield media.Media(mt['filename'])
         self.weblogout()
 
-    def upload_media(self, filename, progress=None):
+    def upload_media(self, filename, progress=None, data=None):
         xid = random.randint(0, 1000000000)
         self._refresh_token()
         uploadthread = webclient.FileUploader(
-            self.wc, '/upload?X-Progress-ID={0}'.format(xid), filename, None)
+            self.wc, '/upload?X-Progress-ID={0}'.format(xid), filename, data)
         uploadthread.start()
         while uploadthread.isAlive():
             uploadthread.join(3)
@@ -1880,6 +1935,7 @@ class XCCClient(IMMClient):
             raise Exception('Unexpected result starting update: %s' % errmsg)
         complete = False
         while not complete:
+            self._refresh_token()
             ipmisession.Session.pause(3)
             rsp = self.wc.grab_json_response(
                 '/api/dataset/imm_firmware_progress')
@@ -2048,7 +2104,7 @@ class XCCClient(IMMClient):
                     '/api/providers/imm_fod', {'FOD_LicenseKeyDelete': licid})
                 break
 
-    def apply_license(self, filename, progress=None):
+    def apply_license(self, filename, progress=None, data=None):
         license_errors = {
             310: "License is for a different model of system",
             311: "License is for a different system serial number",
@@ -2056,7 +2112,8 @@ class XCCClient(IMMClient):
             313: "License is expired",
             314: "License usage limit reached",
         }
-        uploadthread = webclient.FileUploader(self.wc, '/upload', filename)
+        uploadthread = webclient.FileUploader(self.wc, '/upload', filename,
+                                              data=data)
         uploadthread.start()
         uploadthread.join()
         rsp = json.loads(uploadthread.rsp)
