@@ -15,7 +15,7 @@
 import fnmatch
 import struct
 import weakref
-from xml.etree.ElementTree import fromstring
+from xml.etree.ElementTree import fromstring as rfromstring
 import zipfile
 
 import six
@@ -37,6 +37,12 @@ except NameError:
     pass
 
 
+def fromstring(inputdata):
+    if b'!entity' in inputdata.lower():
+        raise Exception('!ENTITY not supported in this interface')
+    return rfromstring(inputdata)
+
+
 def stringtoboolean(originput, name):
     input = originput.lower()
     try:
@@ -76,18 +82,25 @@ def fpc_read_fan_power(ipmicmd):
 
 def fpc_read_psu_fan(ipmicmd, number, sz):
     rsp = ipmicmd.xraw_command(netfn=0x32, command=0xa5, data=(number,))
-    rsp = rsp['data']
-    return struct.unpack_from('<H', rsp[:2])[0]
+    rsp = bytes(rsp['data'])
+    if len(rsp) > 5:
+        return struct.unpack_from('<H', rsp[2:4])[0]
+    else:
+        return struct.unpack_from('<H', rsp[:2])[0]
 
 
 def fpc_get_psustatus(ipmicmd, number, sz):
     rsp = ipmicmd.xraw_command(netfn=0x32, command=0x91)
     mask = 1 << (number - 1)
+    rsp['data'] = bytearray(rsp['data'])
+    if len(rsp['data']) == 10:
+        tmpdata = rsp['data']
+        rsp['data'] = list(struct.unpack('<HHHHBB', tmpdata))
     if len(rsp['data']) == 6:
-        statdata = bytearray([0])
+        statdata = [0]
     else:
-        statdata = bytearray()
-    statdata += bytearray(rsp['data'])
+        statdata = []
+    statdata += rsp['data']
     presence = statdata[3] & mask == mask
     pwrgood = statdata[4] & mask == mask
     throttle = (statdata[6] | statdata[2]) & mask == mask
@@ -123,7 +136,7 @@ def fpc_get_nodeperm(ipmicmd, number, sz):
     perminfo = bytearray(rsp['data'])[1]
     if sz == 6:  # FPC
         permfail = ('\x02', '\x03')
-    elif sz == 2:  # SMM
+    else:  # SMM
         permfail = ('\x02',)
     if perminfo & 0x20:
         if rsp['data'][4] in permfail:
@@ -156,6 +169,11 @@ fpc_sensors = {
         'units': 'W',
         'provider': fpc_read_dc_output,
     },
+    'PSU Power Loss': {
+        'type': 'Power',
+        'units': 'W',
+        'provider': fpc_read_dc_output,
+    },
     'Fan Power': {
         'type': 'Power',
         'units': 'W',
@@ -192,11 +210,14 @@ fpc_sensors = {
 def get_sensor_names(size):
     global fpc_sensors
     for name in fpc_sensors:
-        if size == 2 and name in ('Fan Power', 'Total Power Capacity'):
+        if size != 6 and name in ('Fan Power', 'Total Power Capacity',
+                                  'DC Power'):
+            continue
+        if size == 6 and name == 'PSU Power Loss':
             continue
         sensor = fpc_sensors[name]
         if 'elements' in sensor:
-            for elemidx in range(sensor['elements'] * size):
+            for elemidx in range(sensor['elements'] * (size & 0b11111)):
                 elemidx += 1
                 yield '{0} {1}'.format(name, elemidx)
         else:
@@ -206,11 +227,14 @@ def get_sensor_names(size):
 def get_sensor_descriptions(size):
     global fpc_sensors
     for name in fpc_sensors:
-        if size == 2 and name in ('Fan Power', 'Total Power Capacity'):
+        if size != 6 and name in ('Fan Power', 'Total Power Capacity',
+                                  'DC Power'):
+            continue
+        if size == 6 and name == 'PSU Power Loss':
             continue
         sensor = fpc_sensors[name]
         if 'elements' in sensor:
-            for elemidx in range(sensor['elements'] * size):
+            for elemidx in range(sensor['elements'] * (size & 0b11111)):
                 elemidx += 1
                 yield {'name': '{0} {1}'.format(name, elemidx),
                        'type': sensor['type']}
@@ -222,8 +246,11 @@ def get_fpc_firmware(bmcver, ipmicmd, fpcorsmm):
     mymsg = ipmicmd.xraw_command(netfn=0x32, command=0xa8)
     builddata = bytearray(mymsg['data'])
     name = None
-    if fpcorsmm == 2:  # SMM
-        name = 'SMM'
+    if fpcorsmm != 6:  # SMM
+        if fpcorsmm >> 5:
+            name = 'SMM2'
+        else:
+            name = 'SMM'
         buildid = '{0}{1}{2}{3}{4}{5}{6}'.format(
             *[chr(x) for x in builddata[-7:]])
     elif len(builddata) == 8:
@@ -233,7 +260,12 @@ def get_fpc_firmware(bmcver, ipmicmd, fpcorsmm):
     bmcmajor, bmcminor = [int(x) for x in bmcver.split('.')]
     bmcver = '{0}.{1:02d}'.format(bmcmajor, bmcminor)
     yield (name, {'version': bmcver, 'build': buildid})
-    yield ('PSOC', {'version': '{0}.{1}'.format(builddata[2], builddata[3])})
+    if fpcorsmm == 6:
+        yield ('PSOC', {'version': '{0}.{1}'.format(builddata[2],
+                                                    builddata[3])})
+    else:
+        yield ('PSOC', {'version': '{0}.{1}'.format(builddata[3],
+                                                    builddata[4])})
 
 
 def get_sensor_reading(name, ipmicmd, sz):
@@ -305,19 +337,23 @@ class SMMClient(object):
         accountinfo = fromstring(rspbody)
         for rule in self.rulemap:
             ruleinfo = accountinfo.find(self.rulemap[rule])
-            if ruleinfo:
+            if ruleinfo is not None:
                 settings[rule] = {'value': int(ruleinfo.text)}
-        rsp = self.ipmicmd.xraw_command(0x34, 3)
-        fanmode = self.fanmodes[bytearray(rsp['data'])[0]]
-        settings['fanspeed'] = {
-            'value': fanmode, 'default': 'Normal',
-            'help': ('Adjust the fan speed of the D2 Chassis. Capped settings '
-                     'will reduce fan speed for better acoustic experience at '
-                     'the expense of performance.  Normal is using the Lenovo '
-                     'engineered cooling adjustments across the full range. '
-                     'Boosted adds fanspeed to the Normal response to '
-                     'provide more aggressive cooling.'),
-            'possible': [self.fanmodes[x] for x in self.fanmodes]}
+        dwc = self.ipmicmd.xraw_command(0x32, 0x94)
+        dwc = bytearray(dwc['data'])
+        if len(dwc) != 3 or dwc[0] == 1:
+            rsp = self.ipmicmd.xraw_command(0x34, 3)
+            fanmode = self.fanmodes[bytearray(rsp['data'])[0]]
+            settings['fanspeed'] = {
+                'value': fanmode, 'default': 'Normal',
+                'help': ('Adjust the fan speed of the D2 Chassis. Capped '
+                         'settings will reduce fan speed for better acoustic '
+                         'experience at the expense of performance.  Normal '
+                         'is using the Lenovo engineered cooling adjustments '
+                         'across the full range. Boosted adds fanspeed to '
+                         'the Normal response to provide more aggressive '
+                         'cooling.'),
+                'possible': [self.fanmodes[x] for x in self.fanmodes]}
         powercfg = self.ipmicmd.xraw_command(0x32, 0xa2)
         powercfg = bytearray(powercfg['data'])
         if len(powercfg) == 5:
@@ -472,7 +508,8 @@ class SMMClient(object):
         self.ipmicmd.xraw_command(netfn=0x32, command=0xa4,
                                   data=[int(bay), 2])
 
-    def get_diagnostic_data(self, savefile, progress=None):
+    def get_diagnostic_data(self, savefile, progress=None, autosuffix=False,
+                            variant=None):
         rsp = self.ipmicmd.xraw_command(netfn=0x32, command=0xb1, data=[0])
         if bytearray(rsp['data'])[0] != 0:
             raise Exception("Service data generation already in progress")
@@ -493,7 +530,12 @@ class SMMClient(object):
                 progress({'phase': 'initializing', 'progress': initpct})
         if self.wc is None:
             raise Exception("Failed to connect to web api")
-        url = '/preview/smm-ffdc.tgz?ST1={0}'.format(self.st1)
+        if variant and variant >> 5:
+            url = '/preview/smm2-ffdc.tgz?ST1={0}'.format(self.st1)
+        else:
+            url = '/preview/smm-ffdc.tgz?ST1={0}'.format(self.st1)
+        if autosuffix and not savefile.endswith('.tgz'):
+            savefile += '-smm-ffdc.tgz'
         fd = webclient.FileDownloader(self.wc, url, savefile)
         fd.start()
         while fd.isAlive():
@@ -670,8 +712,15 @@ class SMMClient(object):
     def update_firmware(self, filename, data=None, progress=None, bank=None):
         if progress is None:
             progress = lambda x: True
-        if not data and zipfile.is_zipfile(filename):
+        z = None
+        if data and hasattr(data, 'read'):
+            if zipfile.is_zipfile(data):
+                z = zipfile.ZipFile(data)
+            else:
+                data.seek(0)
+        elif data is None and zipfile.is_zipfile(filename):
             z = zipfile.ZipFile(filename)
+        if z:
             for tmpname in z.namelist():
                 if tmpname.endswith('.rom'):
                     filename = tmpname
