@@ -12,6 +12,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+import os
 import struct
 import time
 try:
@@ -19,6 +20,8 @@ try:
 except ImportError:
     from urllib.parse import urlencode
 
+import six
+
 import pyghmi.exceptions as exc
 import pyghmi.media as media
 import pyghmi.redfish.oem.generic as generic
@@ -32,9 +35,25 @@ class HpmSection(object):
                  'hash_size', 'combo_image']
 
 
-def read_hpm(filename):
+def cstr_to_str(cstr):
+    try:
+        endidx = cstr.index(b'\x00')
+        cstr = cstr[:endidx]
+    except Exception:
+        pass
+    if not isinstance(cstr, str):
+        cstr = cstr.decode('utf8')
+    return cstr
+
+
+def read_hpm(filename, data):
     hpminfo = []
-    with open(filename, 'rb') as hpmfile:
+    if data:
+        hpmfile = data
+        hpmfile.seek(0)
+    else:
+        hpmfile = open(filename, 'rb')
+    try:
         hpmfile.seek(0x20)
         skip = struct.unpack('>H', hpmfile.read(2))[0]
         hpmfile.seek(skip + 1, 1)
@@ -70,13 +89,17 @@ def read_hpm(filename):
         endpos = hpmfile.tell()
         if currpos < (endpos - 512):
             raise Exception("Unexpected end of HPM file")
+    finally:
+        if not data:
+            hpmfile.close()
     return hpminfo
 
 
 class TsmHandler(generic.OEMHandler):
     hostnic = 'usb0'
 
-    def __init__(self, sysinfo, sysurl, webclient, cache=None, fish=None):
+    def __init__(self, sysinfo, sysurl, webclient, cache=None, fish=None,
+                 gpool=None):
         if cache is None:
             cache = {}
         self._wc = None
@@ -85,17 +108,128 @@ class TsmHandler(generic.OEMHandler):
         self.csrftok = None
         self.isipmi = bool(fish)
         self.fish = fish
-        super(TsmHandler, self).__init__(sysinfo, sysurl, webclient, cache)
+        self.fishclient = None
+        super(TsmHandler, self).__init__(sysinfo, sysurl, webclient, cache,
+                                         gpool)
         self.tsm = webclient.thehost
         self._certverify = webclient._certverify
 
+    def get_bmc_configuration(self):
+        wc = self.wc
+        rsp, status = wc.grab_json_response_with_status(
+            '/api/settings/dns-info')
+        if status != 200:
+            raise Exception(repr(rsp))
+        settings = {}
+        settings['dns_domain'] = {
+            'value': rsp['domain_name']
+        }
+        dnssrvs = []
+        for idx in range(3):
+            currsrv = rsp.get('dns_server{0}'.format(idx + 1), '::')
+            if currsrv and currsrv != '::':
+                dnssrvs.append(currsrv)
+        settings['dns_servers'] = {'value': ','.join(dnssrvs)}
+        rsp, status = wc.grab_json_response_with_status(
+            '/api/LockoutPolicystatus')
+        if status == 200:
+            isenabled = rsp.get('Status', 0) == 1
+            if isenabled:
+                settings['password_login_failures'] = {'value': rsp.get(
+                    'Attemptstimes', 0)}
+            else:
+                settings['password_login_failures'] = {'value': 0}
+        rsp, status = wc.grab_json_response_with_status(
+            '/api/GetPWComplex')
+        if status == 200:
+            settings['password_complexity'] = {'value': rsp.get(
+                'pw_complex', 0)}
+        return settings
+
+    def set_bmc_configuration(self, changeset):
+        dnschgs = {}
+        wc = self.wc
+        for key in changeset:
+            if isinstance(changeset[key], six.string_types):
+                changeset[key] = {'value': changeset[key]}
+            currval = changeset[key].get('value', None)
+            if 'dns_servers'.startswith(key.lower()):
+                srvs = currval.split(',')
+                for idx in range(3):
+                    if idx < len(srvs):
+                        dnschgs['dns_server{0}'.format(idx + 1)] = srvs[idx]
+                    else:
+                        dnschgs['dns_server{0}'.format(idx + 1)] = ''
+            if 'dns_domain'.startswith(key.lower()):
+                dnschgs['domain_name'] = currval
+            if 'password_complexity'.startswith(key.lower()):
+                self._set_pass_complexity(currval, wc)
+            if 'password_login_failures'.startswith(key.lower()):
+                self._set_pass_lockout(currval, wc)
+        if dnschgs:
+            self._set_dns_config(dnschgs, wc)
+
+    def _set_pass_complexity(self, currval, wc):
+        rsp, status = wc.grab_json_response_with_status(
+            '/api/SetPWComplex', {'Enable': currval})
+        if status != 200:
+            raise Exception(repr(rsp))
+
+    def _set_pass_lockout(self, currval, wc):
+        rsp, status = wc.grab_json_response_with_status(
+            '/api/LockoutPolicystatus')
+        if status != 200:
+            raise Exception(repr(rsp))
+        request = {
+            'SameStatus': 0,
+            'Lock_min': rsp.get('Locktime', 5),
+            'Rest_min': rsp.get('Resettime', 1),
+            'Attemptstimes': rsp.get('Attemptstimes', 3)
+        }
+        if currval == 0:
+            request['Enable'] = 0
+        else:
+            request['Enable'] = 1
+            request['Attemptstimes'] = currval
+        rsp, status = wc.grab_json_response_with_status(
+            '/api/SetLockoutPolicy', request)
+        if status != 200:
+            raise Exception(repr(rsp))
+
+    def _set_dns_config(self, dnschgs, wc):
+        rsp, status = wc.grab_json_response_with_status(
+            '/api/settings/dns-info')
+        if status != 200:
+            raise Exception(repr(rsp))
+        rsp['domain_manual'] = 1
+        for i in range(3):
+            keyn = 'dns_server{0}'.format(i + 1)
+            if rsp[keyn] == '::':
+                rsp[keyn] = ''
+        for chg in dnschgs:
+            rsp[chg] = dnschgs[chg]
+        rsp, status = wc.grab_json_response_with_status(
+            '/api/settings/dns-info', rsp, method='PUT')
+        if status != 200:
+            raise Exception(repr(rsp))
+        rsp, status = wc.grab_json_response_with_status(
+            '/api/settings/dns/restart', {'dns_status': 1}, method='PUT')
+        if status != 200:
+            raise Exception(repr(rsp))
+
     def clear_uefi_configuration(self):
+        if not self.fishclient:
+            self.init_redfish()
         return self.fishclient.clear_system_configuration()
 
     def get_uefi_configuration(self, hideadvanced=True):
+        if not self.fishclient:
+            self.init_redfish()
         return self.fishclient.get_system_configuration(hideadvanced)
 
     def set_uefi_configuration(self, changeset):
+        if not self.fishclient:
+            self.init_redfish()
         return self.fishclient.set_system_configuration(changeset)
 
     def get_diagnostic_data(self, savefile, progress=None, autosuffix=False):
@@ -136,7 +270,59 @@ class TsmHandler(generic.OEMHandler):
             self.tsm, self.username, self.password,
             verifycallback=self._certverify)
 
-    def get_firmware_inventory(self, components, raisebypass=True):
+    def get_ntp_enabled(self):
+        wc = self.wc
+        rsp, status = wc.grab_json_response_with_status(
+            '/api/settings/date-time')
+        if status != 200:
+            raise Exception(repr(rsp))
+        return rsp.get('ntp_auto_date', 0) > 0
+
+    def set_ntp_enabled(self, enabled):
+        wc = self.wc
+        rsp, status = wc.grab_json_response_with_status(
+            '/api/settings/date-time')
+        if status != 200:
+            raise Exception(repr(rsp))
+        rsp['ntp_auto_date'] = 1 if enabled else 0
+        rsp, status = wc.grab_json_response_with_status(
+            '/api/settings/date-time', rsp, method='PUT')
+        if status != 200:
+            raise Exception(repr(rsp))
+
+    def get_ntp_servers(self):
+        wc = self.wc
+        rsp, status = wc.grab_json_response_with_status(
+            '/api/settings/date-time')
+        if status != 200:
+            raise Exception(repr(rsp))
+        srvs = []
+        pntp = rsp.get('primary_ntp', None)
+        if pntp:
+            srvs.append(pntp)
+        pntp = rsp.get('secondary_ntp', None)
+        if pntp:
+            srvs.append(pntp)
+        return srvs
+
+    def set_ntp_server(self, server, index=0):
+        wc = self.wc
+        rsp, status = wc.grab_json_response_with_status(
+            '/api/settings/date-time')
+        if status != 200:
+            raise Exception(repr(rsp))
+        if index == 0:
+            rsp['primary_ntp'] = server
+        elif index == 1:
+            rsp['secondary_ntp'] = server
+        rsp['ntp_auto_date'] = 1
+        rsp, status = wc.grab_json_response_with_status(
+            '/api/settings/date-time', rsp, method='PUT')
+        if status != 200:
+            raise Exception(repr(rsp))
+
+    def get_firmware_inventory(self, components, raisebypass=True,
+                               ipmicmd=None):
         wc = self.wc
         fwinf, status = wc.grab_json_response_with_status(
             '/api/DeviceVersion')
@@ -170,12 +356,25 @@ class TsmHandler(generic.OEMHandler):
                 lxpmres['version'] = '{0}.{1:02x}'.format(
                     lxpminf['main'], subver)
             yield ('LXPM', lxpmres)
+        if ipmicmd:
+            rsp = ipmicmd.xraw_command(0x3c, 0x40, data=(7, 2))
+            buildid = cstr_to_str(bytes(rsp['data']))
+            yield ('LXPM Windows Driver Bundle', {'build': buildid})
+            rsp = ipmicmd.xraw_command(0x3c, 0x40, data=(7, 3))
+            buildid = cstr_to_str(bytes(rsp['data']))
+            yield ('LXPM Linux Driver Bundle', {'build': buildid})
         name = 'TSM'
         fwinf, status = wc.grab_json_response_with_status('/api/get-sysfwinfo')
         if status != 200:
             raise Exception('Error {0} retrieving TSM version: {1}'.format(
                 status, fwinf))
         for cinf in fwinf:
+            if 'fw_ver' not in cinf:
+                continue
+            if cinf.get('buildname', None) == 'N/A':
+                continue
+            if '.' not in cinf['fw_ver']:
+                continue
             bmcinf = {
                 'version': cinf['fw_ver'],
                 'build': cinf['buildname'],
@@ -222,18 +421,21 @@ class TsmHandler(generic.OEMHandler):
     def update_firmware(self, filename, data=None, progress=None, bank=None):
         wc = self.wc
         wc.set_header('Content-Type', 'application/json')
+        basefilename = os.path.basename(filename)
         if filename.endswith('.hpm'):
-            return self.update_hpm_firmware(filename, progress, wc)
-        elif 'uefi' in filename and filename.endswith('.rom'):
-            return self.update_sys_firmware(filename, progress, wc)
-        elif 'amd-sas' in filename and filename.endswith('.bin'):
-            return self.update_sys_firmware(filename, progress, wc, type='bp')
-        elif 'lxpm' in filename and filename.endswith('.img'):
-            return self.update_lxpm_firmware(filename, progress, wc)
+            return self.update_hpm_firmware(filename, progress, wc, data)
+        elif 'uefi' in basefilename and filename.endswith('.rom'):
+            return self.update_sys_firmware(filename, progress, wc, data=data)
+        elif 'amd-sas' in basefilename and filename.endswith('.bin'):
+            return self.update_sys_firmware(filename, progress, wc, data=data,
+                                            type='bp')
+        elif (('lxpm' in basefilename or 'fw_drv' in basefilename)
+                and filename.endswith('.img')):
+            return self.update_lxpm_firmware(filename, progress, wc, data)
         else:
             raise Exception('Unsupported filename {0}'.format(filename))
 
-    def update_lxpm_firmware(self, filename, progress, wc):
+    def update_lxpm_firmware(self, filename, progress, wc, data):
         hdrs = wc.stdheaders.copy()
         hdrs['Content-Length'] = 0
         rsp = wc.grab_json_response_with_status(
@@ -242,7 +444,7 @@ class TsmHandler(generic.OEMHandler):
         # name fwimage filname filename application/x-raw-disk-image...
         fu = webclient.FileUploader(
             wc, '/api/maintenance/LXPMUpload',
-            filename, formname='fwimage')
+            filename, data, formname='fwimage')
         fu.start()
         while fu.isAlive():
             fu.join(3)
@@ -265,7 +467,8 @@ class TsmHandler(generic.OEMHandler):
             '/api/maintenance/Outofflash', method='PUT', headers=hdrs)
         return 'complete'
 
-    def update_sys_firmware(self, filename, progress, wc, type='uefi'):
+    def update_sys_firmware(self, filename, progress, wc, type='uefi',
+                            data=None):
         if type == 'bp':
             rsp = wc.grab_json_response_with_status('/api/chassis-status')
             if rsp[0]['power_status'] == 1:
@@ -290,7 +493,7 @@ class TsmHandler(generic.OEMHandler):
             '/api/maintenance/{0}'.format(updatemode),
             method='PUT', headers=hdrs)
         fu = webclient.FileUploader(
-            wc, '/api/maintenance/{0}'.format(fileupload), filename,
+            wc, '/api/maintenance/{0}'.format(fileupload), filename, data,
             formname='fwimage')
         fu.start()
         while fu.isAlive():
@@ -337,16 +540,16 @@ class TsmHandler(generic.OEMHandler):
             return 'pending'
         raise Exception('Update Failure')
 
-    def update_hpm_firmware(self, filename, progress, wc):
+    def update_hpm_firmware(self, filename, progress, wc, data):
         rsp = wc.grab_json_response('/api/maintenance/hpm/freemem')
         if 'MemFree' not in rsp:
             raise Exception('System Not Ready for update')
         if filename not in hpm_by_filename:
-            hpminfo = read_hpm(filename)
+            hpminfo = read_hpm(filename, data)
             if len(hpminfo) != 3:
                 raise Exception(
                     'This HPM update is currently not supported')
-            hpm_by_filename[filename] = read_hpm(filename)
+            hpm_by_filename[filename] = read_hpm(filename, data)
         else:
             hpminfo = hpm_by_filename[filename]
         rsp, status = wc.grab_json_response_with_status(
